@@ -8,10 +8,6 @@
 #define INDICATOR_MESSAGES_DBUS_SERVICE_OBJECT     "/org/ayatana/indicator/messages/service"
 #define INDICATOR_MESSAGES_DBUS_SERVICE_INTERFACE  "org.ayatana.indicator.messages.service"
 
-#define LAUNCHER_MENUITEM_TYPE           "launcher-item"
-#define LAUNCHER_MENUITEM_PROP_APP_NAME  "application-name"
-#define LAUNCHER_MENUITEM_PROP_APP_DESC  "application-description"
-
 #define APPLICATION_MENUITEM_TYPE        "application-item"
 #define APPLICATION_MENUITEM_PROP_NAME   "app-name"
 #define APPLICATION_MENUITEM_PROP_COUNT  "app-count"
